/* Copyright (C) 2017-2020 J.F.Dockes
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU Lesser General Public License as published by
 *   the Free Software Foundation; either version 2.1 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU Lesser General Public License for more details.
 *
 *   You should have received a copy of the GNU Lesser General Public License
 *   along with this program; if not, write to the
 *   Free Software Foundation, Inc.,
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */
#ifndef _PROTOCOLINFO_HX_INCLUDED_
#define _PROTOCOLINFO_HX_INCLUDED_

#include <string>
#include <unordered_set>

/**
 * A small class to handle reading/storing/updating the protocolinfo
 * data, which is used by a number of services 
 */
class Protocolinfo {
public:
    static Protocolinfo *the();
    const std::string& gettext();
    const std::unordered_set<std::string>& getsupportedformats();
    bool ok();
private:
    class Internal;
    Internal *m{nullptr};

    Protocolinfo();
    Protocolinfo(const Protocolinfo &);
    Protocolinfo& operator=(const Protocolinfo &);
};

#endif /* _PROTOCOLINFO_HX_INCLUDED_ */
